import numpy as np
import mcfit
from scipy.integrate import quad
from scipy.interpolate import InterpolatedUnivariateSpline as spline
from scipy.special import gamma as Gamma_func
from scipy.special import hyp1f1
from scipy.special import hyperu
from scipy.special import factorial, factorial2


class ZeldovichPowerIntegral(mcfit.mcfit):

    """

    Class to calculate the Zel'dovich power spectrum using mcfit

    """

    def __init__(self, r, n):
        self.n = n
        UK = mcfit.kernels.Mellin_SphericalBesselJ(self.n)
        mcfit.mcfit.__init__(self, r, UK, q=1.5-n, lowring=True)

        # set pre and post factors
        self.prefac = r**(3-n)
        self.postfac = (2*np.pi)**1.5


class PowerIntegral:

    """

    Class to calculate the real and redshift space power spectra using mcfit

    Parameters:

    - n_val is the number of spherical-Bessel functions summed over
    - ki is the k value
    - P is the linear power spectrum function
    - q is the array of q values
    - A is the value of A(z)^2
    - front is the expression at the front of the integral
    - exponent_k_squared is the expression in the exponent multiplied by k^2
    - zero_lag_1 is the zero lag term
    - sigma_psi is the variance at zero displacement
    - kmax is the maximum k value to integrate the low-k approximation to

    """

    def calc_power_rs(self, n_val, ki, P, q, A, front, exponent_k_squared, zero_lag_1, sigma_psi, kmax):

        """
        Function to calculate the real space power spectrum
        """

        Power = 0.0

        f = np.array([], dtype=np.float128)

        # Calculate the low-k approximation if ki<5e-5

        if ki < 5e-3:

            Power = calc_low_k_approx(ki, P, sigma_psi, A, kmax)

        else:

            for n in range(0, int(n_val)):

                I = ZeldovichPowerIntegral(q, n)

                if n > 0:

                    f = (ki*front)**n*np.exp(-0.5*ki**2*exponent_k_squared)

                else:

                    f = np.exp(-0.5*ki**2*exponent_k_squared) - np.exp(-ki**2*zero_lag_1)

                kk, this_Pzel = I(f, extrap=False)

                Power += spline(kk, this_Pzel)(ki)

        return Power


def calc_low_k_approx(k, P, sigma_psi, A, kmax):

    """
    Function to calculate the low k approximation in real space
    """

    Q3 = quad(lambda q: (P(q)/q)**2, 1e-5, kmax)[0]

    term_1 = (1.0-A*k**2*sigma_psi+0.5*A**2*k**4*sigma_psi**2)*A*P(k)
    term_2 = 0.5*(k**4)*Q3/(10.0*np.pi**2)

    return term_1+term_2



import numpy as np
import sys
import classylss
import classylss.binding as CLASS


class Cosmo:

    """
    Class to calculate cosmological functions

    - k_max = the maximum k value used when calculating the power spectra
    - h = H_0/100
    - omega0_b = $\Omega_bh^2$ the baryon density today
    - omega0_cdm = $\Omega_cdmh^2$ the CDM density today
    - n_s = the spectral index
    - gauge = the perturbation gauge used

    - z_val = the redshift value or values to calculate the function at
    - k = the k values to calculate the function at

    """

    def __init__(self, h, omega0_b, omega0_cdm, k_max, n_s, sigma_8, verbose, gauge, output, **kwargs):

        # Read in the cosmological parameters as defined in ctm.py

        self.h = h
        self.omega0_b = omega0_b
        self.omega0_cdm = omega0_cdm
        self.P_k_max = k_max+1.0
        self.n_s = n_s
        self.sigma_8 = sigma_8
        self.gauge = gauge
        self.output = output

        # Initialise Classylss

        self.engine = CLASS.ClassEngine({'h': self.h, 'omega_cdm': self.omega0_cdm, 'omega_b': self.omega0_b, 'n_s': self.n_s, 'gauge': self.gauge,
        'output': self.output, 'sigma8': self.sigma_8, 'P_k_max_1/Mpc': self.P_k_max})

    def scale_factor(self, z_val):

        """
        Function to calculate the scale factor a=(1+z)^{-1}
        """

        return np.power(1.0+z_val, -1)

    def calc_hubble(self, z_val):

        """
        Function to calculate the Hubble function H(z)
        """

        H0 = self.h*100.0

        bg = CLASS.Background(self.engine)

        H = H0*bg.efunc(z_val)

        return H

    def calc_linear_growth(self, z_val):

        """
        Function to calculate the linear growth factor D_1
        """

        bg = CLASS.Background(self.engine)

        D_1 = bg.scale_independent_growth_factor(z_val)

        return D_1

    def calc_independent_linear_growth(self, z_val):

        """
        Function to calculated f=dlnD_1/dlna
        """

        bg = CLASS.Background(self.engine)

        f = bg.scale_independent_growth_rate(z_val)

        return f

    def calc_linear_power(self, k,  z_val=0.0):

        """
        Function to calculate the linear power spectrum
        """

        bg = CLASS.Background(self.engine)

        sp = CLASS.Spectra(self.engine)

        D_1 = bg.scale_independent_growth_factor(z_val)

        pk_lin = D_1**2*sp.get_pklin(k, 0.0)

        return pk_lin

    def Omega_m(self):

        """
        Function to calculate \Omega_m
        """

        return (self.omega0_cdm+self.omega0_b)/self.h**2

    def H0(self):

        """
        Function to calculate H0
        """

        return self.h*100.0



import os
import sys

# Functions used to print progress of various processes


def progress_func_time_dep(func):

    return print("Calculated "+str(func))


def progress_func_power(i, nk):

    if nk <= 1000:

        if i == 0:

            print("Calculating the power spectrum:")

        else:

            n_bar = 10  # size of progress bar

            j = i/nk

            sys.stdout.write('\r')
            sys.stdout.write(f"[{'=' * int(n_bar * j):{n_bar}s}] {int(100 * j)}%")
            sys.stdout.flush()

        if i == nk-1:

            sys.stdout.write('\n')

    if nk > 1000:

        if i == 0:

            print("Calculating the power spectrum:")

        else:

            n_bar = 10  # size of progress bar

            j = i/nk

            sys.stdout.write('\r')
            sys.stdout.write(f"[{'=' * int(n_bar * j):{n_bar}s}] {int(100 * j)}%")
            sys.stdout.flush()


        if i == nk-1:

            sys.stdout.write('\n')    

    return


def progress_func_power_save(filename, i, nk):

    if i == 0:

        print("Started calculating power")

        with open(os.path.join("power_progress_"+filename+".txt"), 'a') as file:

            file.writelines("Started calculating power"+'\n')

    if i % 100 == 0:

        percent = (float(i)/float(nk))*100.0

        print("Completed "+str(percent)+"%")

        with open(os.path.join("power_progress_"+filename+".txt"), 'a') as file:

            file.writelines("Percent complete: "+str(percent)+"%"+'\n')

    return



import numpy as np
import mcfit
from scipy.integrate import quad
from scipy.interpolate import InterpolatedUnivariateSpline as spline
from scipy.special import gamma as Gamma_func
from scipy.special import hyp1f1
from scipy.special import hyperu
from scipy.special import factorial, factorial2


class ZeldovichPowerIntegral(mcfit.mcfit):

    """

    Class to calculate the Zel'dovich power spectrum using mcfit

    """

    def __init__(self, r, n):
        self.n = n
        UK = mcfit.kernels.Mellin_SphericalBesselJ(self.n)
        mcfit.mcfit.__init__(self, r, UK, q=1.5-n, lowring=True)

        # set pre and post factors
        self.prefac = r**(3-n)
        self.postfac = (2*np.pi)**1.5


class PowerIntegral:

    """

    Class to calculate the real and redshift space power spectra using mcfit

    Parameters:

    - n_val is the number of spherical-Bessel functions summed over
    - ki is the k value
    - P is the linear power spectrum function
    - q is the array of q values
    - A is the value of A(z)^2
    - front is the expression at the front of the integral
    - exponent_k_squared is the expression in the exponent multiplied by k^2
    - zero_lag_1 is the zero lag term
    - sigma_psi is the variance at zero displacement
    - kmax is the maximum k value to integrate the low-k approximation to

    """

    def calc_power_rs(self, n_val, ki, P, q, A, front, exponent_k_squared, zero_lag_1, sigma_psi, kmax):

        # Function to calculate the real space power spectrum

        Power = 0.0

        f = np.array([], dtype=np.float128)

        # Calculate the low-k approximation if ki<5e-5

        if ki < 5e-3:

            Power = calc_low_k_approx(ki, P, sigma_psi, A, kmax)

        else:

            for n in range(0, int(n_val)):

                I = ZeldovichPowerIntegral(q, n)

                if n > 0:

                    f = (ki*front)**n*np.exp(-0.5*ki**2*exponent_k_squared)

                else:

                    f = np.exp(-0.5*ki**2*exponent_k_squared) - np.exp(-ki**2*zero_lag_1)

                kk, this_Pzel = I(f, extrap=False)

                Power += spline(kk, this_Pzel)(ki)

        return Power

    def calc_power_rsd(self, ki, q, P, front, exponent_k_squared, zero_lag_1, kmax, alpha_0, alpha_1, C, xi, kappa, gamma, sigma, rsd_exponent_1, rsd_exponent_2, sigma_psi, A, f_val, mu_k):

        # Function to calculate the redshift space power spectrum

        Power = 0.0

        # Calculate the low k approximation if ki<5e-4

        if ki < 5e-5:

            Power = calc_low_k_approx_rsd(ki, P, sigma_psi, A, kmax, f_val, mu_k)

        else:
            for n in range(0, 32):

                K_n = 0.0
                I = ZeldovichPowerIntegral(q, n)

                if n > 0:

                    for l in range(0, 1):

                        K_n += np.exp(-0.5*ki**2*rsd_exponent_1-0.5*ki**2*rsd_exponent_2)*calc_F_l(l, alpha_0, alpha_1, -0.5*ki**2*C, xi, kappa, gamma, sigma)*(-1)**l*np.power(alpha_1, n)*np.power(kappa, n-l)*np.power(gamma**2/sigma, l)*hyperu(-l, n-l+1.0, 0.5*ki**2*C*kappa*alpha_1)


                    f = (ki*front)**n*np.exp(-0.5*ki**2*exponent_k_squared)*K_n

                else:

                    for l in range(0, 1):

                        K_n += np.exp(-0.5*ki**2*rsd_exponent_1-0.5*ki**2*rsd_exponent_2)*calc_F_l(l,alpha_0,alpha_1,-0.5*ki**2*C,xi,kappa,gamma,sigma)*(-1)**l*np.power(kappa,-l)*np.power(gamma**2/sigma,l)*hyperu(-l,-l+1.0,0.5*ki**2*C*kappa*alpha_1)


                    f = (np.exp(-0.5*ki**2*exponent_k_squared) - np.exp(-ki**2*zero_lag_1))*K_n

                kk, this_Pzel = I(f, extrap=False)

                Power += spline(kk, this_Pzel)(ki)

            return Power

def calc_low_k_approx(k, P, sigma_psi, A, kmax):

    # Low k approximation in real space

    Q3 = quad(lambda q: (P(q)/q)**2, 1e-5, kmax)[0]

    term_1 = (1.0-A*k**2*sigma_psi+0.5*A**2*k**4*sigma_psi**2)*A*P(k)
    term_2 = 0.5*(k**4)*Q3/(10.0*np.pi**2)

    return term_1+term_2


def calc_low_k_approx_rsd(k, P, sigma_psi, A, kmax, f_val, mu_k):

    # Low k approximation in redshift space

    factor = f_val*(f_val+2.0)*mu_k**2
    Q3 = quad(lambda q: (P(q)/q)**2, 1e-5, kmax)[0]

    term_1 = (1.0-A*k**2*sigma_psi-A*k**2*factor*sigma_psi+0.5*A**2*k**4*sigma_psi**2)*A*(1.0+f_val*mu_k**2)**2*P(k)+(1.0-A*k**2*sigma_psi-A*k**2*factor*sigma_psi)*((1.0+f_val*mu_k**2)**2*P(k)*k**2*(1.0+factor)*sigma_psi)
    term_2 = 0.5*(k**4)*Q3/(10.0*np.pi**2)

    return term_1+term_2


def calc_F_l(l, alpha_0, alpha_1, C, xi, kappa, gamma, sigma):

    # Calculate the function F_l for the redshift space power spectrum

    F_l = (np.power(np.pi, -0.5)*np.power(4.0, l)*Gamma_func(0.5))/(Gamma_func(1.0)*Gamma_func(1.0-l)*Gamma_func(2.0*l+1.0))*hyp1f1(l, l+0.5, C*alpha_1*(gamma**2/sigma))*hyp1f1(0.5, 1.0, C*alpha_0*(xi*gamma**2/(sigma*kappa)))

    for m in range(1, l):

        F_l += ((np.power(np.pi, -0.5)*np.power(-1.0, m)*np.power(4.0, l-m)*Gamma_func(m+0.5))/(Gamma_func(m+1.0)*Gamma_func(1.0+2.0*m-l)*Gamma_func(2.0*l-2.0*m+1.0)))*np.power(alpha_0/alpha_1, m)*np.power(xi/kappa, m)*hyp1f1(l-2.0*m, l-m+0.5, C*alpha_1*(gamma**2/sigma))*hyp1f1(m+0.5, m+1.0, C*alpha_0*(xi*gamma**2/(sigma*kappa)))

    return np.array(F_l, dtype=np.float128)



import numpy as np
from .cosmology.cosmology import Cosmo
from .power_spectrum.LPT import LPTPower
from .power_spectrum.ctm_power import PowerSpec
from .power_spectrum.correlation_funcs import Corr

class CTM:

    """

    Class to calculate the CTM power spectrum

    Cosmological Parameters:

    - min_k = the minimum k value,
    - max_k = the maximum k value,
    - nk = the number of k values,
    - h = H_0/100,
    - omega0_b = $\Omega_bh^2$ the baryon density today,
    - omega0_cdm = $\Omega_cdmh^2$ the CDM density today,
    - n_s = the spectral index

    Zel'dovich Parameters:

    - input_k = input k values (if you do not specify then automatic k values are returned)
    - input_P = input P values at z=0 you must also specify the k values used to calculate input_P
    - z_val = the redshift value at which the Zel'dovich power spectrum will be calculated
    - k_c = cutoff k value if using an initial Gaussian damped power spectrum
    - n_val = the number of spherical Bessel functions summed over (n_val=32.)

    CTM Parameters:

    - k_c = cutoff k value if using an initial Gaussian damped power spectrum
    - input_k = input k values (if you do not specify then automatic k values are returned)
    - input_P = input P values at zinit you must also specify the k values used to calculate input_P as input_k_init. Note if your initial z is not z_init=100 then you must also specify z_init unless you have passed input_A
    - input_k_init = input k values at which the input_P power spectrum is calculated at
    - z_val = the redshift value at which the CTM power spectrum will be calculated
    - epsilon = the expansion parameter which controls the size of the second-order CTM terms (epsilon=1)
    - z_init = the initial redshift value the time dependent factors are integrated from (z_init=100)
    - input_z = input z values
    - input_A = input A values you must also specify the z values used to calculate input_A
    - input_B = input B values you must also specify the z values used to calculate input_B

    Correlation function parameters:

    - min_r = the minimum r value used to calculate the correlation function
    - max_r = the maximum r value used to calculate the correlation function
    - nr = the number of r values used to calculate the correlation function

    """

    def __init__(self, min_k=1e-5, max_k=100.0, nk=3000, h=0.6737, omega0_b=0.02233, omega0_cdm=0.11933, n_s=0.9665, sigma_8=0.8102, verbose=False, gauge='sync', output='mPk', **kwargs):

        self.min_k = min_k
        self.max_k = max_k
        self.nk = nk
        self.h = h
        self.omega0_b = omega0_b
        self.omega0_cdm = omega0_cdm
        self.n_s = n_s
        self.sigma_8 = sigma_8
        self.verbose = verbose
        self.gauge = gauge
        self.output = output

    def linear_growth_factor(self, z_val=0.0):

        """
        Function to calculate the linear growth factor using Classylss
        """

        return Cosmo(h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, k_max=self.max_k, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).calc_linear_growth(z_val=z_val)

    def linear_power(self, input_k, z_val=0.0):

        """
        Function to calculate the linear power spectrum using Classylss
        """

        return Cosmo(h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, k_max=self.max_k, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).calc_linear_power(input_k, z_val=z_val)

    def zeldovich_power(self, n_val=32, z_val=0.0, kc=0.0, input_k=np.zeros(10), input_P=np.zeros(10), save=False):

        """
        Function to calculate the Zel'dovich power spectrum in real space.
        """

        return LPTPower(min_k=self.min_k, max_k=self.max_k, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).calc_zeldovich_power(n_val=n_val, z_val=z_val, kc=kc, input_k=input_k, input_P=input_P, save=save)

    def ctm_power(self,  n_val=32, zinit=100.0, z_val=0.0, epsilon=1.0, save=False, kc=0.0, input_k=np.zeros(10), input_P=np.zeros(10), input_k_init=np.zeros(10), input_z=np.zeros(10), input_A=np.zeros(10), input_B=np.zeros(10)):

        """
        Function to calculate the CTM power spectrum in real space.

        If no A and B functions are specified the Beyond Zel'dovich approximation is used. See the Documentation for details.
        """
        return PowerSpec(min_k=self.min_k, max_k=10.0, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).calc_ctm_power(n_val=n_val, zinit=zinit, z_val=z_val, epsilon=epsilon, save=save, kc=kc, input_k=input_k, input_P=input_P, input_k_init=input_k_init, input_z=input_z, input_A=input_A, input_B=input_B)

    def corr_func(self, k_values, P_values, min_r=1.0, max_r=1000.0, nr=10000):

        """
        Function to calculate the two-point correlation function given k values and power spectrum values
        """

        return Corr(min_k=self.min_k, max_k=self.max_k, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).corr_func(min_r=min_r, max_r=max_r, nr=nr, k_values=k_values, P_values=P_values)

    def corr_func_zel(self, min_r=1.0, max_r=1000.0, nr=10000, n_val=32, z_val=0.0, kc=0.0, input_k=np.zeros(10), input_P=np.zeros(10), save=False):

        """
        Function to calculate the Zel'dovich two-point correlation function in real space.
        """

        return Corr(min_k=self.min_k, max_k=self.max_k, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).zeldovich_corr_func(min_r=min_r, max_r=max_r, nr=nr, z_val=z_val, kc=kc, input_k=input_k, input_P=input_P, save=save)

    def corr_func_ctm(self, min_r=1.0, max_r=1000.0, nr=10000, n_val=32, zinit=100.0, z_val=0.0, epsilon=1.0, save=False, kc=0.0, input_k=np.zeros(10), input_P=np.zeros(10), input_k_init=np.zeros(10), input_z=np.zeros(10), input_A=np.zeros(10), input_B=np.zeros(10)):

        """
        Function to calculate the CTM two-point correlation function in real space

        If no A and B functions are specified the Beyond Zel'dovich approximation is used. See the Documentation for details.
        """

        return Corr(min_k=self.min_k, max_k=self.max_k, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).ctm_corr_func(min_r=min_r, max_r=max_r, nr=nr, n_val=n_val, zinit=zinit, z_val=z_val, epsilon=epsilon, save=save, kc=kc, input_k=input_k, input_P=input_P, input_k_init=input_k_init, input_z=input_z, input_A=input_A, input_B=input_B)



import numpy as np
import sys
from scipy.interpolate import interp1d as interp
from scipy.integrate import trapz
import mcfit
from mcfit import SphericalBessel as sph
import warnings
import os
from ..cosmology.cosmology import Cosmo
from ..cosmology.time_dependence import TimeDep
from .integration import PowerIntegral
from .covariances import calc_covariances_func
from ..misc.timer import Timer
from ..misc.progress_update import progress_func_power
from ..misc.progress_update import progress_func_power_save

warnings.filterwarnings("ignore")


class LPTPower:

    """

    Class to calculate the Zel'dovich power spectrum

    Parameters:

    - min_k is the minimum k value
    - max_k is the maximum k value
    - nk is the number of k values
    - h is the Hubble constant
    - omega0_b is the baryon density today
    - omega0_cdm is the CDM density today
    - n_s is the spectral index
    - k_max is the maximum k value used when calculating the power spectrum

    - z_val is the redshift value at which the Zel'dovich power spectrum is calculated
    - input_k are user specified input k values (if you do not specify then automatic k values are returned)
    - input_P is an input power spectrum at z=0 if this is given then the k values used to calculate input_P must also be given as input_k
    - k_c is the cutoff k value if using an initial Gaussian damped power spectrum
    - n_val is the number of spherical Bessel functions summed over

    """

    def __init__(self, min_k, max_k, nk, h, omega0_b, omega0_cdm, n_s, sigma_8, verbose, gauge, output, **kwargs):

        self.h = h
        self.omega0_b = omega0_b
        self.omega0_cdm = omega0_cdm
        self.n_s = n_s
        self.sigma_8 = sigma_8
        self.gauge = gauge
        self.output = output
        self.verbose = verbose
        self.nk = nk
        self.min_k = min_k
        self.max_k = max_k

        # Define the k vector for the integrals

        self.k = np.logspace(np.log10(self.min_k), np.log10(self.max_k), self.nk)

        # Initialise Classylss

        self.cosmo = Cosmo(self.h, self.omega0_b, self.omega0_cdm, self.max_k+1.0, self.n_s, self.sigma_8, self.verbose, self.gauge, self.output)

    def calc_zeldovich_power(self, n_val=32, z_val=0.0, kc=0.0, input_k=np.zeros(10), input_P=np.zeros(10), save=False):

        """
        Function to calculate the power spectrum for the Zel'dovich approximation using the method presented in 1209.0780 and 1410.1617
        """

        time = Timer()
        time.start()

        # Calculate the linear power spectrum for the specified redshift value

        if input_k.all() != 0.0 and input_P.all() != 0.0:

            D_1 = self.cosmo.calc_linear_growth(z_val)

            if kc != 0.0:

                P_vals = np.exp(-(input_k/kc)**2)*(D_1)**2*input_P

            else:

                P_vals = (D_1)**2*input_P

            P_func = interp(input_k, P_vals)

        else:

            P = self.cosmo.calc_linear_power(self.k, 0.0)

            D_1 = self.cosmo.calc_linear_growth(z_val)

            if kc != 0.0:

                P_vals = np.exp(-(self.k/kc)**2)*(D_1)**2*P

            else:

                P_vals = (D_1)**2*P

            P_func = interp(self.k, P_vals)

            k_int = self.k

            nk_calc = self.nk

            max_k_calc = self.max_k

        print("Calculated the input power spectrum")

        # Calculate the covariances

        sigma_psi, q_vals, X_vals, Y_vals, eta_E, sigma_0, D_vals, F_vals, G_vals = calc_covariances_func(k_int, P_func)

        print("Calculated the covariances")

        XY = X_vals + Y_vals

        front = Y_vals
        exponent_k_squared = XY
        zero_lag_1 = sigma_psi

        A = 1.0

        # Begin calculating the Zel'dovich power spectrum

        P_calculated = np.zeros_like(k_int)

        for i in range(nk_calc):

            progress_func_power(i, nk_calc)

            P_calculated[i] = PowerIntegral().calc_power_rs(n_val, k_int[i], P_func, q_vals, A, front, exponent_k_squared, zero_lag_1, sigma_psi, max_k_calc)

        time.stop()

        if input_k.all() != 0.0:

            P_calculated_func = interp(k_int, P_calculated)
            P_return = P_calculated_func(input_k)

            if save is True:

                for i in range(len(P_return)):

                    with open(os.path.join("P_zeldovich_"+str(int(z_val))+".txt"), 'a') as file:

                        file.writelines(str(P_return[i])+'\n')

            return P_return

        else:

            if save is True:

                for i in range(len(P_calculated)):

                    with open(os.path.join("P_zeldovich_"+str(int(z_val))+".txt"), 'a') as file:

                        file.writelines(str(P_calculated[i])+'\n')

            return k_int, P_calculated



import numpy as np
import sys
from scipy.interpolate import interp1d as interp
import warnings
import os
from ..cosmology.cosmology import Cosmo
from ..cosmology.time_dependence import TimeDep
from .integration import PowerIntegral
from .covariances import calc_covariances_func
from ..misc.timer import Timer
from ..misc.progress_update import progress_func_power
from scipy.interpolate import InterpolatedUnivariateSpline as interpolate

warnings.filterwarnings("ignore")


class PowerSpec:

    """

    Class to calculate the CTM power spectrum

    Parameters:

    - min_k = the minimum k value,
    - max_k = the maximum k value,
    - nk = the number of k values,
    - h = H_0/100,
    - omega0_b = $\Omega_bh^2$ the baryon density today,
    - omega0_cdm = $\Omega_cdmh^2$ the CDM density today,
    - n_s = the spectral index

    - k_c = cutoff k value if using an initial Gaussian damped power spectrum
    - input_k are user specified input k values (if you do not specify then automatic k values are returned)
    - input_P is an input power spectrum at z=0 if this is given then the k values used to calculate input_P must also be given as input_k
    - z_val = the redshift value at which the CTM power spectrum will be calculated
    - epsilon = the expansion parameter which controls the size of the second-order CTM term
    - z_init = the initial redshift value the time dependent factors are integrated from
    - input_z = input z values
    - input_A = input A values you must also specify the z values used to calculate input_A
    - input_B = input B values you must also specify the z values used to calculate input_B
    - n_val is the number of spherical Bessel functions summed over

    """

    def __init__(self, min_k, max_k, nk, h, omega0_b, omega0_cdm, n_s, sigma_8, verbose, gauge, output, **kwargs):

        self.h = h
        self.omega0_b = omega0_b
        self.omega0_cdm = omega0_cdm
        self.n_s = n_s
        self.sigma_8 = sigma_8
        self.gauge = gauge
        self.output = output
        self.verbose = verbose
        self.nk = nk
        self.min_k = min_k
        self.max_k = max_k

        # Define the k vector for the integrals

        self.k_int = np.logspace(np.log10(self.min_k), np.log10(self.max_k), self.nk)

        # Initialise classylss

        self.cosmo = Cosmo(self.h, self.omega0_b, self.omega0_cdm, self.max_k+1.0, self.n_s, self.sigma_8, self.verbose, self.gauge, self.output)

    def calc_ctm_power(self, n_val=32., zinit=100.0, z_val=0.0, epsilon=1.0, save=False, kc=0.0, input_k=np.zeros(10), input_k_init=np.zeros(10), input_P=np.zeros(10), input_z=np.zeros(10), input_A=np.zeros(10), input_B=np.zeros(10)):

        """
        Function to calculate the CTM power spectrum
        """

        time = Timer()
        time.start()

        if input_P.all() != 0.0:

            if kc != 0.0:

                P_vals = np.exp(-(self.k_int/kc)**2)*input_P

            else:

                P_vals = input_P

            P_func = interp(input_k_init, P_vals)

        else:

            P = self.cosmo.calc_linear_power(self.k_int)

            D_1 = self.cosmo.calc_linear_growth(zinit)

            if kc != 0.0:

                P_vals = np.exp(-(self.k_int/kc)**2)*(D_1)**2*P

            else:

                P_vals = (D_1)**2*P

            P_func = interp(self.k_int, P_vals)

        print("Calculated the input power spectrum")

        if input_A.all() != 0.0 and input_B.all() == 0.0:

            A_func = interp(input_z, input_A)
            A_squared_val = A_func(z_val)**2

            B_squared_val = TimeDep(zinit, self.h, self.omega0_b, self.omega0_cdm, self.max_k+1.0, self.n_s, self.sigma_8, self.verbose, self.gauge, self.output).calc_B(input_A, z_val)

        if input_A.all() != 0.0 and input_B.all() != 0.0:

            A_func = interp(input_z, input_A)
            A_squared_val = A_func(z_val)**2

            B_func = interp(input_z, input_B)
            B_squared_val = B_func(z_val)**2

        if input_A.all() == 0.0 and input_B.all() == 0.0:

            D_1_val = self.cosmo.calc_linear_growth(z_val)
            D_1_init = self.cosmo.calc_linear_growth(zinit)

            A_squared_val = (D_1_val/D_1_init)**2

            w = 1.5*self.cosmo.Omega_m()*self.cosmo.H0()**2
            zeta_val = TimeDep(zinit, self.h, self.omega0_b, self.omega0_cdm, self.max_k+1.0, self.n_s, self.sigma_8, self.verbose, self.gauge, self.output).calc_zeta(z_val)
            B_squared_val = (epsilon*w*zeta_val)**2

        print('Calculated time dependent functions A(z) and B(z)')

        sigma_psi, q_vals, X_vals, Y_vals, eta_E, sigma_0, D_vals, F_vals, G_vals = calc_covariances_func(self.k_int, P_func)

        print("Calculated the covariances")

        XY = X_vals + Y_vals

        W_prime = -(1.0/3.0)*eta_E*sigma_psi+(sigma_psi-0.5*X_vals)*((2.0/3.0)*D_vals-(1.0/9.0)*sigma_0)

        Z_prime = (1.0/9.0)*G_vals**2+(2.0/3.0)*F_vals*(sigma_psi-0.5*X_vals)+Y_vals*((1.0/18.0)*sigma_0-(1.0/3.0)*D_vals-(1.0/3.0)*F_vals)

        exponent_1 = A_squared_val*X_vals-2.0*B_squared_val*W_prime
        exponent_3 = A_squared_val*Y_vals-2.0*B_squared_val*Z_prime

        front = exponent_3
        exponent_k_squared = exponent_1+exponent_3
        zero_lag_1 = sigma_psi*(A_squared_val+(1.0/3.0)*B_squared_val*eta_E)

        P_calculated = np.zeros_like(self.k_int)

        for i in range(self.nk):

            progress_func_power(i, self.nk)

            P_calculated[i] = PowerIntegral().calc_power_rs(n_val, self.k_int[i], P_func, q_vals, A_squared_val, front, exponent_k_squared, zero_lag_1, sigma_psi, self.max_k)

        time.stop()

        if input_k.all() != 0.0:

            P_calculated_func = interp(self.k_int, P_calculated)
            P_return = P_calculated_func(input_k)

            if save is True:

                for i in range(len(P_return)):

                    with open(os.path.join("results/P_gctm_"+str(int(z_val))+".txt"), 'a') as file:

                        file.writelines(str(P_return[i])+'\n')

            return P_return

        else:

            if save is True:

                for i in range(self.nk):

                    with open(os.path.join("results/P_gctm_"+str(int(z_val))+".txt"), 'a') as file:

                        file.writelines(str(P_calculated[i])+'\n')

            return self.k_int, P_calculated



import numpy as np
import matplotlib.pyplot as plt

import matplotlib as mpl

from matplotlib import animation
from IPython.display import HTML

from matplotlib import rc
from matplotlib import gridspec
from mpl_toolkits.axes_grid1 import AxesGrid
from matplotlib.transforms import blended_transform_factory

mpl.rcParams['text.latex.preamble'] = [r"\usepackage{amsmath}"]
mpl.rcParams["text.usetex"] = True
mpl.rcParams['font.family'] = 'sans-serif'
mpl.rcParams['font.sans-serif'] = 'cm'
mpl.rcParams["lines.linewidth"] = 2.2
mpl.rcParams["axes.linewidth"] = 1.5
mpl.rcParams["axes.labelsize"] = 14.
mpl.rcParams["xtick.top"] = True
mpl.rcParams["xtick.labelsize"] = 14.
mpl.rcParams["xtick.direction"] = "in"
mpl.rcParams["ytick.right"] = True
mpl.rcParams["ytick.labelsize"] = 14.
mpl.rcParams["ytick.direction"] = "in"
mpl.rcParams["xtick.minor.bottom"] = False
mpl.rcParams["xtick.minor.top"] = False
mpl.rcParams["ytick.minor.left"] = False
mpl.rcParams["ytick.minor.right"] = False

colour_1 = "dodgerblue"
colour_2 = "#EE442F"
colour_3 = "forestgreen"
colour_4 = "purple"
colour_5 = "deeppink"

colours = ["black", colour_1, colour_2, colour_3, colour_4, colour_5]
linestyles = ["-", "--", "-.", ":"]

class Plotting(object):

        def single_panel_figure(self, x_data, y_data, line_labels, x_label, y_label, x_min, x_max, y_min, y_max, type, legend=True, loc="upper left", fontsize=12., save=False, filename="power_spec.pdf"):

            fig = plt.figure(figsize=(6, 5))
            gs = gridspec.GridSpec(1, 1)

            gs.update(wspace=0.05, hspace=0.1)

            ax1 = fig.add_subplot(gs[0])

            if type == "loglog":

                for i in range(len(x_data)):

                    ax1.loglog(x_data[i], y_data[i], label=line_labels[i], color=colours[i], linestyle=linestyles[i])
                    ax1.set_xlim([x_min, x_max])
                    ax1.set_ylim([y_min, y_max])
                    ax1.set_xlabel(x_label)
                    ax1.set_ylabel(y_label)

            if type == "semilogx":

                for i in range(len(x_data)):

                    ax1.semilogx(x_data[i], y_data[i], label=line_labels[i], color=colours[i], linestyle=linestyles[i])
                    ax1.set_xlim([x_min, x_max])
                    ax1.set_ylim([y_min, y_max])
                    ax1.set_xlabel(x_label)
                    ax1.set_ylabel(y_label)

            if type == "semilogy":

                for i in range(len(x_data)):

                    ax1.semilogy(x_data[i], y_data[i], label=line_labels[i], color=colours[i], linestyle=linestyles[i])
                    ax1.set_xlim([x_min, x_max])
                    ax1.set_ylim([y_min, y_max])
                    ax1.set_xlabel(x_label)
                    ax1.set_ylabel(y_label)

            if type == "plot":

                for i in range(len(x_data)):

                    ax1.plot(x_data[i], y_data[i], label=line_labels[i], color=colours[i], linestyle=linestyles[i])
                    ax1.set_xlim([x_min, x_max])
                    ax1.set_ylim([y_min, y_max])
                    ax1.set_xlabel(x_label)
                    ax1.set_ylabel(y_label)

            if legend == True:

                ax1.legend(loc=loc, frameon=False, fontsize=fontsize)

            if save == True:

                plt.savefig(filename, bbox_inches="tight")

        def two_panel_figure(self, x_data_1, y_data_1, x_data_1, y_data_1, line_labels, x_label, y_label, x_min, x_max, y_min, y_max, type, legend=True, loc="upper left", fontsize=12., save=False, filename="power_spec.pdf"):

            fig = plt.figure(figsize=(6, 5))
            gs = gridspec.GridSpec(1, 1)

            gs.update(wspace=0.05, hspace=0.1)

            ax1 = fig.add_subplot(gs[0])

            if type == "loglog":

                for i in range(len(x_data)):

                    ax1.loglog(x_data[i], y_data[i], label=line_labels[i], color=colours[i], linestyle=linestyles[i])
                    ax1.set_xlim([x_min, x_max])
                    ax1.set_ylim([y_min, y_max])
                    ax1.set_xlabel(x_label)
                    ax1.set_ylabel(y_label)

            if type == "semilogx":

                for i in range(len(x_data)):

                    ax1.semilogx(x_data[i], y_data[i], label=line_labels[i], color=colours[i], linestyle=linestyles[i])
                    ax1.set_xlim([x_min, x_max])
                    ax1.set_ylim([y_min, y_max])
                    ax1.set_xlabel(x_label)
                    ax1.set_ylabel(y_label)

            if type == "semilogy":

                for i in range(len(x_data)):

                    ax1.semilogy(x_data[i], y_data[i], label=line_labels[i], color=colours[i], linestyle=linestyles[i])
                    ax1.set_xlim([x_min, x_max])
                    ax1.set_ylim([y_min, y_max])
                    ax1.set_xlabel(x_label)
                    ax1.set_ylabel(y_label)

            if type == "plot":

                for i in range(len(x_data)):

                    ax1.plot(x_data[i], y_data[i], label=line_labels[i], color=colours[i], linestyle=linestyles[i])
                    ax1.set_xlim([x_min, x_max])
                    ax1.set_ylim([y_min, y_max])
                    ax1.set_xlabel(x_label)
                    ax1.set_ylabel(y_label)

            if legend == True:

                ax1.legend(loc=loc, frameon=False, fontsize=fontsize)

            if save == True:

                plt.savefig(filename, bbox_inches="tight")



import numpy as np
import sys
from scipy.interpolate import interp1d as interp
from scipy.integrate import trapz
import mcfit
from mcfit import SphericalBessel as sph
import warnings
import os
from ..cosmology.cosmology import Cosmo
from ..cosmology.time_dependence import TimeDep
from .integration import PowerIntegral
from .covariances import calc_covariances_func
from ..misc.timer import Timer
from ..misc.progress_update import progress_func_power
from ..misc.progress_update import progress_func_power_save

warnings.filterwarnings("ignore")


class LPTPower:

    """

    Class the calculate the Zel'dovich power spectrum

    Parameters:

    - min_k is the minimum k value
    - max_k is the maximum k value
    - nk is the number of k values
    - h is the Hubble constant
    - omega0_b is the baryon density today
    - omega0_cdm is the CDM density today
    - n_s is the spectral index
    - k_max is the maximum k value used when calculating the power spectrum


    """

    def __init__(self, min_k, max_k, min_k_zel, max_k_zel, nk, h, omega0_b, omega0_cdm, n_s, sigma_8, verbose, gauge, output, **kwargs):

        self.h = h
        self.omega0_b = omega0_b
        self.omega0_cdm = omega0_cdm
        self.n_s = n_s
        self.sigma_8 = sigma_8
        self.gauge = gauge
        self.output = output
        self.verbose = verbose
        self.nk = nk
        self.min_k = min_k
        self.max_k = max_k
        self.min_k_zel = min_k_zel
        self.max_k_zel = max_k_zel

        # Define the k vector for loop integrals

        self.k = np.logspace(np.log10(self.min_k), np.log10(self.max_k), self.nk)

        self.k_zel = np.logspace(np.log10(self.min_k_zel), np.log10(self.max_k_zel), self.nk)

        # Initialise Classylss

        self.cosmo = Cosmo(self.h, self.omega0_b, self.omega0_cdm, self.max_k+1.0, self.n_s, self.sigma_8, self.verbose, self.gauge, self.output)

    def calc_zeldovich_power(self, n_val=32, z_val=0.0, kc=0.0, input_k=np.zeros(10), input_P=np.zeros(10), save=False):

        # Function to calculate the power spectrum for the Zel'dovich approximation using the method presented in 1209.0780 and 1410.1617

        time = Timer()
        time.start()

        # Calculate the linear power spectrum for the specified redshift value

        if input_k.all() != 0.0 and input_P.all() != 0.0:

            D_1 = self.cosmo.calc_linear_growth(zinit)

            if kc != 0.0:

                P_vals = np.exp(-(input_k/kc)**2)*(D_1)**2*input_P

            else:

                P_vals = (D_1)**2*input_P

            P_func = interp(input_k, P_vals)

        else:

            P = self.cosmo.calc_linear_power(self.k_zel, 0.0)

            D_1 = self.cosmo.calc_linear_growth(z_val)

            if kc != 0.0:

                P_vals = np.exp(-(self.k_zel/kc)**2)*(D_1)**2*P

            else:

                P_vals = (D_1)**2*P

            P_func = interp(self.k_zel, P_vals)

            k_calc = self.k_zel

            nk_calc = self.nk

            max_k_calc = self.max_k_zel

        print("Calculated the input power spectrum")

        # Calculate the covariances

        sigma_psi, q_vals, X_vals, Y_vals, eta_E, sigma_0, D_vals, F_vals, G_vals = calc_covariances_func(k_calc, P_func)

        print("Calculated the covariances")

        XY = X_vals + Y_vals

        front = Y_vals
        exponent_k_squared = XY
        zero_lag_1 = sigma_psi

        A = 1.0

        # Begin calculating the Zel'dovich power spectrum

        P_calculated = np.zeros_like(k_calc)

        for i in range(nk_calc):

            progress_func_power(i, nk_calc)

            P_calculated[i] = PowerIntegral().calc_power_rs(n_val, k_calc[i], P_func, q_vals, A, front, exponent_k_squared, zero_lag_1, sigma_psi, max_k_calc)

        time.stop()

        if input_k.all() != 0.0:

            P_calculated_func = interp(k_calc, P_calculated)
            P_return = P_calculated_func(input_k)

            if save is True:

                for i in range(len(P_return)):

                    with open(os.path.join("P_zeldovich_"+str(int(z_val))+".txt"), 'a') as file:

                        file.writelines(str(P_return[i])+'\n')

            return P_return

        else:

            if save is True:

                for i in range(len(P_calculated)):

                    with open(os.path.join("P_zeldovich_"+str(int(z_val))+".txt"), 'a') as file:

                        file.writelines(str(P_calculated[i])+'\n')

            return k_calc, P_calculated

    def one_loop(self, z_val=0.0, input_k=np.zeros(10), input_P=np.zeros(10), save=False):

        # Function to calculate the power spectrum for LPT+1-loop using the method presented in 1209.0780 and 1410.1617

        # Load weight parameters for the Q and R integrals
        full_path = os.path.realpath(__file__)
        glxval, glwval = np.loadtxt(os.path.dirname(full_path)+"/gl_128.txt", unpack=True)
        glxval = glxval.reshape(1, -1)
        glwval = glwval.reshape(1, -1)

        # Calculate the linear power spectrum for the loop integrals

        P_lin = self.cosmo.calc_linear_power(self.k)
        P_int = interp(self.k, P_lin)

        time = Timer()
        time.start()

        if input_k.all() != 0.0 and input_P.all() != 0.0:

            P_vals = input_P

            P_func = interp(input_k, P_vals)

            k_calc = input_k

            nk_calc = len(input_k)

            max_k_calc = max(input_k)

        else:

            P = self.cosmo.calc_linear_power(self.k_zel, 0.0)

            D_1 = self.cosmo.calc_linear_growth(z_val)

            P_vals = (D_1)**2*P

            P_func = interp(self.k_zel, P_vals)

            k_calc = self.k_zel

            nk_calc = self.nk

            max_k_calc = self.max_k_zel

        print("Calculated the input power spectrum")

        # Calculate the Q and R functions as defined in 1209.0780

        k_loop = np.logspace(-4, np.log10(50), self.nk)

        Q = np.zeros_like(k_loop)
        R = np.zeros_like(k_loop)

        for i in range(len(k_calc)):

            Q[i] = Q_external(k_loop[i], P_int, glxval, glwval)
            R[i] = R_external(k_loop[i], P_int, glxval, glwval)

        Q_func = interp(k_loop, Q)
        R_func = interp(k_loop, R)

        sigma_psi, sigma_22, sigma_13, q_vals, X_loop, Y_loop = calc_covariances_func_one_loop(k_loop, P_func, Q_func, R_func)

        print("Calculated the covariances")

        XY = X_loop + Y_loop

        # Define the arrays to be passed to the integral Class (see documentation for more details)

        front = Y_loop
        exponent_k_squared = XY
        zero_lag_1 = sigma_psi + sigma_22 + 2.0*sigma_13

        # Define the time dependent factor A(z) which is 1 in the case of LPT

        A = 1.0

        # Begin calculating the LPT 1-loop power spectrum

        P_calculated = np.zeros_like(k_calc)

        for i in range(nk_calc):

            progress_func_power(i, nk_calc)

            P_calculated[i] = PowerIntegral().calc_power_rs(k_calc[i], P_func, q_vals, A, front, exponent_k_squared, zero_lag_1, sigma_psi, max_k_calc)

        time.stop()

        if input_k.all() != 0.0:

            P_calculated_func = interp(k_calc, P_calculated)
            P_return = P_calculated_func(input_k)

            if save is True:

                for i in range(len(P_return)):

                    with open(os.path.join("P_lpt_oneloop_"+str(int(z_val))+".txt"), 'a') as file:

                        file.writelines(str(P_return[i])+'\n')

            return P_return

        else:

            if save is True:

                for i in range(len(P_calculated)):

                    with open(os.path.join("P_lpt_oneloop_"+str(int(z_val))+".txt"), 'a') as file:

                        file.writelines(str(P_calculated[i])+'\n')

            return self.k, P_calculated


def Q1(r, x):

    # Function to calculate the Q1 function defined in equation (B17) in Arxiv:1209.0780 for LPT 1-loop

    y = 1.0+r**2-2.0*r*x
    Q_1 = r**2*(1.0-x**2)**2/y**2
    return Q_1


def Q_internal(k, r, P, glxval, glwval):

    # Function to calculate the internal integral in equation (B17) in Arxiv:1209.0780 for LPT 1-loop

    def func(r, x):

        return P(k*np.sqrt(1.0+r**2-2.0*r*x))*Q1(r, x)

    return (glwval*func(r, glxval)).sum(axis=-1)


def Q_external(k, P, glxval, glwval):

    # Function to calculate the external integral in equation (B17) in Arxiv:1209.0780 for LPT 1-loop

    fac = k**3/(2.0*np.pi)**2
    kint = np.logspace(-5, 2, 1000)
    r = (kint/k)
    tol = 1e-5
    absr1 = abs(r-1)
    mask = absr1 < tol

    y = np.zeros_like(r)
    y[~mask] = Q_internal(k, r[~mask].reshape(-1, 1), P, glxval, glwval)
    y *= P(kint)

    return fac*trapz(y, r)


def R1(r, x):

    # Function to calculate the R1 function defined in equation (B16) in Arxiv:1209.0780 for LPT 1-loop

    y = 1.0+r**2-2.0*r*x

    return (r**2*(1.0-x**2)**2)/y


def R_internal(r, P, glxval, glwval):

    # Function to calculate the internal integral defined in equation (B16) in Arxiv:1209.0780 for LPT 1-loop

    def func(r, x):

        return R1(r, x)

    return (glwval*func(r, glxval)).sum(axis=-1)


def R_external(k, P, glxval, glwval):

    # Function to calculate the external integral defined in equation (B16) in Arxiv:1209.0780 for LPT 1-loop

    fac = k**3/(2.0*np.pi)**2*P(k)
    kint = np.logspace(-4, np.log10(5e2), 1000)
    r = (kint/k)
    absr1 = abs(r-1)
    tol = 1e-5
    mask = absr1 < tol

    y = np.zeros_like(r)
    y[~mask] = R_internal(r[~mask].reshape(-1, 1), P, glxval, glwval)
    y *= P(kint)

    return fac*trapz(y, r)



import numpy as np
import mcfit
from mcfit import SphericalBessel as sph
from scipy.integrate import quad
from scipy.interpolate import InterpolatedUnivariateSpline as spline
from scipy.special import spherical_jn
from scipy.interpolate import interp1d as interp
from scipy.misc import derivative
import sys


# Constant definitions

npi2 = np.power(np.pi, -2)
renorm = np.sqrt(0.5*np.pi)


def calc_covariances_func(k, P):

    # Function to calculate the covariances X, Y, U, V, R, S, T
    # for an input power spectrum

    sigma_psi = calc_sigma_psi(k, P)
    q_vals, X_vals, Y_vals = calc_X_and_Y(k, P, sigma_psi)
    eta_E = calc_eta(k, P)
    _, D_vals, F_vals, sigma_0 = calc_D_and_F(k, P)
    _, G_vals = calc_G(k, P)

    return sigma_psi, q_vals, X_vals, Y_vals, eta_E, sigma_0, D_vals, F_vals, G_vals


def calc_covariances_func_one_loop(k, P, Q, R):

    # Function to calculate the covariances X, Y for LPT+1loop power spectrum

    sigma_psi = calc_sigma_psi(k, P)
    sigma_22 = calc_sigma_22(k, Q)
    sigma_13 = calc_sigma_13(k, R)
    q_vals, X_lin, Y_lin = calc_X_and_Y(k, P, sigma_psi)
    X22, Y22 = calc_X22_and_Y22(k, Q, sigma_22)
    X13, Y13 = calc_X13_and_Y13(k, R, sigma_13)

    X_loop = X_lin + X22 + 2.0*X13
    Y_loop = Y_lin + Y22 + 2.0*Y13

    return sigma_psi, sigma_22, sigma_13, q_vals, X_loop, Y_loop

# Function definitions


def dosph(n, x, f, a, tilt=1.5):

    # Function to calculate spherical Bessel integrals returns q values and integral values
    # Parameters: n = order of Bessel, x = variable of the function, f = function integrating over, a = power law

    func = renorm*np.power(x, a)*f(x)

    return sph(x, nu=n, q=tilt, lowring=True)(func, extrap=True)


def calc_sigma_psi(k, P):

    # Function to calculate the variance of the displacement field \sigma^2{\psi} = \frac{1}{6\pi^2}\int{dkP_L\left(k)}

    q, I0 = dosph(0, k, P, -2)
    sigma_vals = npi2*I0/6.0

    sigma_func = interp(q, sigma_vals)
    sigma_0 = sigma_func(min(q))

    return sigma_0


def calc_X_and_Y(k, P, sigma):

    # Function to calculate X and Y where X = \frac{1}{2\pi^2}\int{dk[\frac{2}{3}-\frac{j_1(kq)}{kq}]P_L} and Y = \frac{1}{2\pi^2}\int{dk[6\frac{j_1(kq)}{kq}-j_0(kq)]P_L} and sigma is \sigma^2_{\psi}

    q, I0 = dosph(0, k, P, -2)
    _, I2 = dosph(2, k, P, -2)

    X = 2.0*sigma - npi2*(1.0/3.0)*(I0+I2)
    Y = npi2*I2

    return q, X, Y


def calc_eta(k, P):

    # Function to calculate eta^2_E = \frac{1}{6\pi^2}\int{dk}k^2P_L

    q, I0 = dosph(0, k, P, 0)
    eta_vals = npi2*I0/6.0

    eta_func = interp(q, eta_vals)
    eta_0 = eta_func(min(q))

    return eta_0

def calc_D_and_F(k, P):

    # Function to calculate D=\frac{1}{2\pi^2}\int{dk kj_3(kq)P_L} and F=-\frac{1}{2\pi^2}\int{dk kj_2(kq)/kqP_L} and sigma^2_0=\frac{1}{\left(2\pi\right)^3}\int{dk}e^{-ik\cdotq}P_L

    q, I0 = dosph(0, k, P, 0)
    _, I2 = dosph(2, k, P, 0)

    F = -0.5*npi2*I2
    D = npi2*(I0+I2)/(6.0)
    sigma_0 = 0.5*npi2*I0

    return q, D, F, sigma_0

def calc_G(k, P):

    q, I1 = dosph(1, k, P, -1)

    G = -0.5*npi2*I1

    return q, G


def calc_sigma_22(k, Q):

    # Function to calculate the variance of the displacement field \sigma^2{\psi} = \frac{1}{6\pi^2}\int{dkQ_1\left(k)}

    q, I0 = dosph(0, k, Q, -2)

    sigma_func = interp(q, npi2*I0/6.0)
    sigma_0 = sigma_func(min(q))

    return (9.0/98.0)*sigma_0


def calc_sigma_13(k, R):

    # Function to calculate the variance of the displacement field \sigma^2{\psi} = \frac{1}{6\pi^2}\int{dkR_1\left(k)}

    q, I0 = dosph(0, k, R, -2)

    sigma_func = interp(q, npi2*I0/6.0)
    sigma_0 = sigma_func(min(q))

    return (5.0/21.0)*sigma_0


def calc_X11_and_Y11(k, P, sigma):

    # Function to calculate X_11 and Y_11 where X_11 = \frac{1}{2\pi^2}\int{dk[\frac{2}{3}-\frac{j_1(kq)}{kq}]P_L} and Y_11 = \frac{1}{2\pi^2}\int{dk[6\frac{j_1(kq)}{kq}-j_0(kq)]P_L} and sigma is \sigma^2_{\psi}

    q, I0 = dosph(0, k, P, -2)
    _, I2 = dosph(2, k, P, -2)

    X11 = 2.0*sigma - npi2*(1.0/3.0)*(I0+I2)
    Y11 = npi2*I2

    return X11, Y11


def calc_X22_and_Y22(k, Q, sigma):

    # Function to calculate X_22 and Y_22 where X_22 = \frac{1}{2\pi^2}\int{dk[\frac{2}{3}-\frac{j_1(kq)}{kq}]Q_1} and Y_22 = \frac{1}{2\pi^2}\int{dk[6\frac{j_1(kq)}{kq}-j_0(kq)]Q_1} and sigma is \sigma^2_{22}

    q, I0 = dosph(0, k, Q, -2)
    _, I2 = dosph(2, k, Q, -2)

    X22 = 2.0*sigma-(9.0/98.0)*(npi2*(1.0/3.0)*(I0+I2))
    Y22 = (9.0/98.0)*npi2*I2

    return X22, Y22


def calc_X13_and_Y13(k, R, sigma):

    # Function to calculate X_13 and Y_13 where X_13 = \frac{1}{2\pi^2}\int{dk[\frac{2}{3}-\frac{j_1(kq)}{kq}]R_1} and Y_13 = \frac{1}{2\pi^2}\int{dk[6\frac{j_1(kq)}{kq}-j_0(kq)]R_1} and sigma is \sigma^2_{13}

    q, I0 = dosph(0, k, R, -2)
    _, I2 = dosph(2, k, R, -2)

    X13 = 2.0*sigma-(5.0/21.0)*(npi2*(1.0/3.0)*(I0+I2))
    Y13 = (5.0/21.0)*npi2*I2

    return X13, Y13



import numpy as np
import sys
from scipy.integrate import quad
from scipy.integrate import dblquad
from scipy.misc import derivative
from scipy.interpolate import interp1d as interp
from .cosmology import Cosmo
import warnings
warnings.filterwarnings("ignore")


class TimeDep:

    """
    Class to calculate the time dependent functions

    Parameters:

    - h is the Hubble constant
    - omega0_b is the baryon fraction today
    - omega0_cdm is the CDM fraction today
    - k_max is the maximum k value used when calculating the power spectra
    - n_s is the spectral index
    - gauge = the perturbation gauge used
    - z_init = the initial redshift defined in the CTM

    - k = the k values to calculate the function at
    - z = the redshift value or values to calculate the time dependent functions at

    """

    def __init__(self, zinit=100.0, h=0.6737, omega0_b=0.02233, omega0_cdm=0.11933, k_max=10.0, n_s=0.9665, sigma_8=0.8102, verbose=False, gauge='sync', output='mPk', **kwargs):

        # Define the initial z value

        self.zinit = zinit

        # Read in the cosmological parameters as specified in gctm.py

        self.h = h
        self.omega0_b = omega0_b
        self.omega0_cdm = omega0_cdm
        self.k_max = k_max+1.0
        self.n_s = n_s
        self.sigma_8 = sigma_8
        self.gauge = gauge
        self.output = output
        self.verbose = verbose

        # Initialise classylss

        self.cosmo = Cosmo(self.h, self.omega0_b, self.omega0_cdm, self.k_max, self.n_s, self.sigma_8, self.verbose, self.gauge, self.output)

    def calc_zeta(self, z):

        """
        Function to calculate \zeta (see documentation for more details)
        """

        if type(z) == np.ndarray:

            # Calculate the Hubble parameter for the redshifts

            hubble_values = self.cosmo.calc_hubble(z)
            H = interp(z, hubble_values)

            # Calculate the linear growth factor D_1 values for the redshifts

            linear_growth_values = self.cosmo.calc_linear_growth(z)
            D_1 = interp(z, linear_growth_values)

            # Calculate the scale factor values for the redshifts

            scale_factor_values = self.cosmo.scale_factor(z)
            scale_factor = interp(z, scale_factor_values)

            # Calculate \zeta

            zeta = np.zeros_like(z)

            def super_conformal(z_prime, z_val):

                return quad(lambda p: np.power(H(p), -1)*(D_1(p)/D_1(self.zinit))**2, self.zinit, z_prime)[0]

            for i in range(len(z)):

                zeta[i] = quad(lambda z_prime: super_conformal(z_prime, z[i])*np.power(scale_factor(z_prime)*H(z_prime),-1), self.zinit, z[i], epsabs=1e-4)[0]

        else:

            # Define the redshifts

            z_vals = np.linspace(0.0, 200.0, 1000)

            # Calculate the Hubble parameter for the redshifts

            hubble_values = self.cosmo.calc_hubble(z_vals)
            H = interp(z_vals, hubble_values)

            # Calculate the linear growth factor D_1 values for the redshifts

            linear_growth_values = self.cosmo.calc_linear_growth(z_vals)
            D_1 = interp(z_vals, linear_growth_values)

            # Calculate the scale factor values for the redshifts

            scale_factor = interp(z_vals, self.cosmo.scale_factor(z_vals))

            # Calculate \zeta

            def super_conformal(z_prime, z_val):

                return quad(lambda p: np.power(H(p), -1)*(D_1(p)/D_1(self.zinit))**2, self.zinit, z_prime)[0]

            zeta = quad(lambda z_prime: super_conformal(z_prime, z)*np.power(scale_factor(z_prime)*H(z_prime),-1), self.zinit, z, epsabs=1e-4)[0]

        return zeta

    def calc_B(self, A_vals, z):

        """
        Function to calculate a general B(z) (see documentation for more details)
        """

        if type(z) == np.ndarray:

            # Calculate the Hubble parameter for the redshifts

            hubble_values = self.cosmo.calc_hubble(z)
            H = interp(z, hubble_values)

            # Calculate the linear growth function D_1 for the redshifts

            linear_growth_values = self.cosmo.calc_linear_growth(z)
            A = interp(z, A_vals)

            # Calculate the scale factor values for the redshifts

            scale_factor_values = self.cosmo.scale_factor(z)
            scale_factor = interp(z, scale_factor_values)

            # Calculate B

            def super_conformal(z_prime, z_val):

                return quad(lambda p: np.power(H(p), -1)*A(p)**2, self.zinit, z_prime)[0]

            B = np.zeros_like(z)

            for i in range(int(len(z))):

                B[i] = quad(lambda z_prime: super_conformal(z_prime, z[i])*np.power(scale_factor(z_prime)*H(z_prime),-1), self.zinit, z[i], epsabs=1e-4)[0]

        else:

            # Define the redshifts

            z_vals = np.linspace(0.0, 200.0, len(A_vals))

            # Calculate the Hubble parameter for the redshifts

            hubble_values = self.cosmo.calc_hubble(z_vals)
            H = interp(z_vals, hubble_values)

            # Calculate the linear growth function D_1 for the redshifts

            linear_growth_values = self.cosmo.calc_linear_growth(z_vals)

            # Interpolate the array of A(z) values

            A = interp(z_vals, A_vals)

            # Calculate the scale factor for the redshifts

            scale_factor = interp(z_vals, self.cosmo.scale_factor(z_vals))

            # Calculate B(z)

            def super_conformal(z_prime, z_val):

                return quad(lambda p: np.power(H(p), -1)*A(p)**2, self.zinit, z_prime)[0]

            B = quad(lambda z_prime: super_conformal(z_prime, z)*np.power(scale_factor(z_prime)*H(z_prime),-1), self.zinit, z, epsabs=1e-4)[0]

        return B



from setuptools import setup

setup(name='cosmotm',
      version='1.11',
      description='Module to compute the power spectrum and two-point correlation function using the CTM',
      url='https://ctm-module.readthedocs.io',
      author='Fran Lane',
      author_email='ctmcodehelp@gmail.com',
      license='MIT',
      packages=['ctm', 'ctm.cosmology', 'ctm.misc', 'ctm.power_spectrum'],
      install_requires=['mcfit', 'classylss', 'scipy'])



import numpy as np
import sys
from scipy.interpolate import interp1d as interp
from scipy.integrate import trapz
import mcfit
from mcfit import SphericalBessel as sph
import warnings
import os
from ..cosmology.cosmology import Cosmo
from ..cosmology.time_dependence import TimeDep
from .LPT import LPTPower
from .ctm_power import PowerSpec
from ..misc.timer import Timer
from ..misc.progress_update import progress_func_power
from ..misc.progress_update import progress_func_power_save

warnings.filterwarnings("ignore")

### Constant definitions

npi2 = np.power(np.pi,-2)
renorm = np.sqrt(0.5*np.pi)


class Corr:

    """

    Class to calculate the two-point correlation functions for the Zel'dovich approximation and second-order CTM

    Parameters:

    - min_k is the minimum k value
    - max_k is the maximum k value
    - nk is the number of k values
    - h is the Hubble constant
    - omega0_b is the baryon density today
    - omega0_cdm is the CDM density today
    - n_s is the spectral index
    - k_max is the maximum k value used when calculating the power spectrum

    - z_val is the redshift value at which the Zel'dovich power spectrum is calculated
    - input_k are user specified input k values (if you do not specify then automatic k values are returned)
    - input_P is an input power spectrum at z=0 if this is given then the k values used to calculate input_P must also be given as input_k
    - k_c is the cutoff k value if using an initial Gaussian damped power spectrum
    - n_val is the number of spherical Bessel functions summed over
    - epsilon is the expansion parameter which controls the size of the second-order CTM term
    - z_init is the initial redshift value the time dependent factors are integrated from
    - input_z is input z values
    - input_A is input A values you must also specify the z values used to calculate input_A
    - input_B is input B values you must also specify the z values used to calculate input_B

    - min_r is the minimum r value used in the calculation of the correlation function
    - max_r is the maximum r value used in the calculation of the correlation function
    - nr is the number of r values used in the calcultion of the correlation function

    """

    def __init__(self, min_k, max_k, nk, h, omega0_b, omega0_cdm, n_s, sigma_8, verbose, gauge, output, **kwargs):

        self.h = h
        self.omega0_b = omega0_b
        self.omega0_cdm = omega0_cdm
        self.n_s = n_s
        self.sigma_8 = sigma_8
        self.gauge = gauge
        self.output = output
        self.verbose = verbose
        self.nk = nk
        self.min_k = min_k
        self.max_k = max_k

        # Define the k vector for the integrals

        self.k_int = np.logspace(np.log10(self.min_k), np.log10(self.max_k), self.nk)

        # Initialise classylss

        self.cosmo = Cosmo(self.h, self.omega0_b, self.omega0_cdm, self.max_k+1.0, self.n_s, self.sigma_8, self.verbose, self.gauge, self.output)

    def corr_func(self, k_values, P_values, min_r=1.0, max_r=1000.0, nr=10000):

        """
        Function to calculate the two-point correlation function given k_values and P_values
        """

        P_func = interp(k_values, P_values)

        new_k_vals = np.logspace(np.log10(1.0/max_r) , np.log10(1.0/min_r), nr)

        r_vals, I0 = dosph(0, new_k_vals, P_func, 0)

        corr_vals = 0.5*npi2*I0

        return r_vals, corr_vals

    def zeldovich_corr_func(self, min_r=1.0, max_r=1000.0, nr=10000, n_val=32, z_val=0.0, kc=0.0, input_k=np.zeros(10), input_P=np.zeros(10), save=False):

        """
        Function to calculate the two-point correlation function for the Zel'dovich approximation
        """

        if input_k.all() == 0.0:

            k_values, P_values = LPTPower(min_k=self.min_k, max_k=self.max_k, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).calc_zeldovich_power(n_val=n_val, z_val=z_val, kc=kc, input_k=input_k, input_P=input_P, save=save)

        else:

            k_values = input_k

            P_values = LPTPower(min_k=self.min_k, max_k=self.max_k, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).calc_zeldovich_power(n_val=n_val, z_val=z_val, kc=kc, input_k=input_k, input_P=input_P, save=save)

        P_func = interp(k_values, P_values)

        new_k_vals = np.logspace(np.log10(1.0/max_r) , np.log10(1.0/min_r), nr)

        r_vals, I0 = dosph(0, new_k_vals, P_func, 0)

        corr_vals = 0.5*npi2*I0

        return r_vals, corr_vals

    def ctm_corr_func(self, min_r=1.0, max_r=1000.0, nr=10000, n_val=32, zinit=100.0, z_val=0.0, epsilon=1.0, save=False, kc=0.0, input_k=np.zeros(10), input_P=np.zeros(10), input_k_init=np.zeros(10), input_z=np.zeros(10), input_A=np.zeros(10), input_B=np.zeros(10)):

        """
        Function to calculate the two-point correlation function for the CTM
        """

        if input_k.all() == 0.0:

            k_values, P_values = PowerSpec(min_k=self.min_k, max_k=10.0, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).calc_ctm_power(n_val=n_val, zinit=zinit, z_val=z_val, epsilon=epsilon, save=save, kc=kc, input_k=input_k, input_P=input_P, input_k_init=input_k_init, input_z=input_z, input_A=input_A, input_B=input_B)

        else:

            k_values = input_k

            P_values = PowerSpec(min_k=self.min_k, max_k=10.0, nk=self.nk, h=self.h, omega0_b=self.omega0_b, omega0_cdm=self.omega0_cdm, n_s=self.n_s, sigma_8=self.sigma_8, verbose=self.verbose, gauge=self.gauge, output=self.output).calc_ctm_power(n_val=n_val, zinit=zinit, z_val=z_val, epsilon=epsilon, save=save, kc=kc, input_k=input_k, input_P=input_P, input_k_init=input_k_init, input_z=input_z, input_A=input_A, input_B=input_B)

        P_func = interp(k_values, P_values)

        new_k_vals = np.logspace(np.log10(1.0/max_r) , np.log10(1.0/min_r), nr)

        r_vals, I0 = dosph(0, new_k_vals, P_func, 0)

        corr_vals = 0.5*npi2*I0

        return r_vals, corr_vals

### Function to calculate spherical Bessel integrals returns q values and integral values
### Parameters: n = order of Bessel, x = variable of the function, f = function integrating over, a = power law

def dosph(n, x, f, a, tilt=1.5):

    func = renorm*np.power(x, a)*f(x)

    return sph(x, nu=n, q=tilt, lowring=True)(func, extrap=True)
